# coding: utf-8
# file generated by setuptools_scm
# don't change, don't track in version control
version = '0.1.5.post2.dev2+gb159f13.d20210528'
version_tuple = (0, 1, 5, 'post2', 'dev2+gb159f13', 'd20210528')
